// Due to incompatiblities between Windows and GNU libraries this file
// can be compiled in three versions depending on preprocessor constants

#if defined(WIN32)
#pragma warning( disable : 4786 )
#endif

#include <iostream>
#include <strstream>
#include <cstdio>
#include "SSMdlReader.h"
#include "SSQModel.h"
#include "SSTypes.h"

void SSMdlReader::pushback()
{
  pushbackflag = 1;
}

SSMdlReader::status SSMdlReader::readLine(ifstream &in)
{
  if(pushbackflag) {
    pushbackflag = 0;
    return pushbackstat;
  }
  do {
    in.getline(buf, 1024);
    lineno++;
  } while(in.good() && (buf[0]=='\0' || buf[0]=='#'));
  if(in.eof()) {
    pushbackstat = IO_EOF;
    return pushbackstat;
  } if(in.fail()) {
    pushbackstat = IO_FAIL;
    return pushbackstat;
  }
  pushbackstat = IO_GOOD;
  return pushbackstat;
}

void SSMdlReader::error(string msg, int io)
{
  if(io)
    cerr << "I/O error on line " << lineno << ":\n" << msg << endl;
  else
    cerr << "Error:\n" << msg << endl;
  exit(EXIT_FAILURE);
}

void SSMdlReader::readVertices(ifstream &in)
{
  while(true) {
    stat = readLine(in);
    if(stat) {
      pushback();
      return;
    }

#if defined(WIN32)
    string test(buf);
    if(test == "vertices" || test == "faces"
       || test == "connections" || test == "knotspacings")
      {
        modedata = test;
        pushback();
        return;
      }

    istrstream is(buf);
    is >> numdata >> coorddata1 >> coorddata2 >> coorddata3;
#elif defined(__SCAN_BUG_WORKAROUND)
    int n = sscanf(buf, "%d %le %le %le", &numdata, &coorddata1, &coorddata2, &coorddata3);
    if(n != 4) {
      pushback();
      return;
    }
#else
    istrstream is(buf);
    is >> numdata >> coorddata1 >> coorddata2 >> coorddata3;
    if(is.fail() || is.eof()) {
      is.clear();
      is >> modedata;
      if(is.fail()) {
        error(string("Illegal file format.\n"
                     "Vertices must be specified as a line\n"
                     "  <vno> <x> <y> <z>\n"
                     "where <vno> is the vertex number, and <x> <y> <z> \n"
                     "are the coordinates"), 1);
      }
      pushback();
      return;
    }
#endif
    vertices.insert(make_pair(numdata, SSVertex(coorddata1, coorddata2, coorddata3)));
  }
}

void SSMdlReader::readFaces(ifstream &in)
{
  map<int, vector<int> >::iterator it;
  int count;
  while(true) {
    count = 0;
    stat = readLine(in);
    if(stat) {
      pushback();
      return;
    }
    istrstream is(buf);
#if defined(WIN32)
    string test(buf);
    if(test == "vertices" || test == "faces"
       || test == "connections" || test == "knotspacings")
      {
        modedata = test;
        pushback();
        return;
      }

    is >> numdata;
#elif defined(__SCAN_BUG_WORKAROUND)
    int n = sscanf(buf, "%i", &numdata);
    if(n != 1) {
      pushback();
      return;
    }
#else
    is >> numdata;
    if(is.fail()) {
      is.clear();
      is >> modedata;
      if(is.fail())
        error("Faces must be specified as a line\n"
              "  <fno> <vno> <vno> <vno> [<vno> ...]\n"
              "where <fno> is the face number, and <vno> is a "
              "valid vertex number.\n", 1);
      pushback();
      return;
    }
#endif
    
#if defined(WIN32)
    while(is >> vnumdata)
      {
#elif defined(__SCAN_BUG_WORKAROUND)
    char *pos = buf + strspn(buf, " ");
    pos = pos + strcspn(pos, " ");
    while(true) {
      int n = sscanf(pos, "%i", &vnumdata);
      if(n!=1) {
        break;
      }
      pos = pos + strspn(pos, " ");
      pos = pos + strcspn(pos, " ");
#else
    while(true) {
      is >> vnumdata;
      if(is.eof()) {
        if(count < 3)
          error("Faces must be specified as a line\n"
                "  <fno> <vno> <vno> <vno> [<vno> ...]\n"
                "where <fno> is the face number, and <vno> is a "
                "valid vertex number.\n", 1);
        break;
      }
      if(is.fail()) {
        error("Faces must be specified as a line\n"
              "  <fno> <vno> <vno> <vno> [<vno> ...]\n"
              "where <fno> is the face number, and <vno> is a "
              "valid vertex number.\n", 1);
      }
#endif
      it = facedata.find(numdata);
      if(it == facedata.end()) {
        vector<int> v;
        v.push_back(vnumdata);
        facedata.insert(make_pair(numdata, v));
      } else {
        it->second.push_back(vnumdata);
      }
#if !defined(WIN32)
      count++;
#endif
    }
  }
}

void SSMdlReader::readConnections(ifstream &in)
{
  while(true) {
    stat = readLine(in);
    if(stat) {
      pushback();
      return;
    }
    istrstream is(buf);
#if defined(WIN32)
    string test(buf);
    if(test == "vertices" || test == "faces"
       || test == "connections" || test == "knotspacings")
      {
        modedata = test;
        pushback();
        return;
      }
    
    is >> numdata >> vnumdata;
#elif defined(__SCAN_BUG_WORKAROUND)
    int n = sscanf(buf, "%d %d", &numdata, &vnumdata);
    if(n != 2) {
      pushback();
      return;
    }
#else
    is >> numdata >> vnumdata;
    if(is.fail()) {
      is.clear();
      is >> modedata;
      if(is.fail()) {
        error("Connections must be specified as a line\n"
              "  <fno> <fno>\n"
              "where <fno> is a valid face number.\n", 1);
      }
      pushback();
      return;
    }
#endif
    connections.push_back(make_pair(numdata, vnumdata));
  }
}
  
void SSMdlReader::readKnotSpacings(ifstream &in)
{
  map<int,vector<double> >::iterator it;
  while(true) {
    stat = readLine(in);
    if(stat) {
      pushback();
      return;
    }
    istrstream is(buf);
#if defined(WIN32)
    string test(buf);
    if(test == "vertices" || test == "faces" 
       || test == "connections" || test == "knotspacings")
      {
        modedata = test;
        pushback();
        return;
      }
    
    is >> numdata;
#elif defined(__SCAN_BUG_WORKAROUND)
    int n = sscanf(buf, "%d", &numdata);
    if(n!=1) {
      pushback();
      return;
    }
#else
    is >> numdata;
    if(is.fail()) {
      is.clear();
      is >> modedata;
      if(is.fail()) {
        error("Knot spacings must be specified as a line\n"
              "  <fno> <ksp> <ksp> [<ksp> ...]\n"
              "where <fno> is a valid face number and <ksp> is a "
              "knot spacing.\n", 1);
      }
      pushback();
      return;
    }
#endif
#if defined(WIN32)
    while(is >> coorddata1 >> coorddata2)
      {
#elif defined(__SCAN_BUG_WORKAROUND)
    char *pos = buf + strspn(buf, " ");
    pos = pos + strcspn(pos, " ");
        
    while(true) {
      int n = sscanf(pos, "%le %le", &coorddata1, &coorddata2);
      if(n!=2) {
        break;
      }
      pos = pos + strspn(pos, " ");
      pos = pos + strcspn(pos, " ");
      pos = pos + strspn(pos, " ");
      pos = pos + strcspn(pos, " ");
#else
    while(true) {
      is >> coorddata1 >> coorddata2;
      if(is.eof() || is.fail()) {
        if(is.eof())
          break;
        is.clear();
        is >> modedata;
        if(is.fail()) {
          error("Knot spacings must be specified as a line\n"
                "  <fno> <ksp> <ksp> [<ksp> ...]\n"
                "where <fno> is a valid face number and <ksp> is a knot spacing.\n"
                "There must be two knot spacing values for each edge in the face.\n", 1);
        }
        pushback();
        return;
      }
#endif
      it = knotspacings.find(numdata);
      if(it == knotspacings.end()) {
        vector<double> v;
        v.push_back(coorddata1);
        v.push_back(coorddata2);
        knotspacings.insert(make_pair(numdata, v));
      } else {
        it->second.push_back(coorddata1);
        it->second.push_back(coorddata2);
      }
    }
  }
}

pair<int, int> SSMdlReader::findEdgeNo(const vector<int> &vec1, 
                                       const vector<int> &vec2, 
                                       int f1, 
                                       int f2)
{
  ostrstream ost;
  int n1 = vec1.size(), n2 = vec2.size(), c11 = 0, c12, c21, c22;  
  while(c11 < n1) 
    {
      c12 = 0;
      while(c12 < n2) 
        {
          if(vec1[c11] == vec2[c12])
            goto cont;
          c12++;
        }
      c11++;
    }
  ost << "Trying to connect faces " << f1 << " and " << f2 
      << " which share no vertices.\n";
  error(ost.str(), 0);
 cont:
  c21 = c11+1 % n1;
  c22 = c12 == 0 ? n2-1 : c12-1;
  if(vec1[c21] == vec2[c22])
    return make_pair(c21, c12);
  c21 = c11 == 0 ? n1-1 : c11-1;
  c22 = c12+1 % n2;
  if(vec1[c21] == vec2[c22])
    return make_pair(c11, c22);
  ost << "Trying to connect faces " << f1 << " and " << f2
      << " which share no edge.\nMaybe one of the faces has the wrong orientation.\n";
  error(ost.str(), 0);
  return make_pair(0,0); // This statement is never reached
}

void SSMdlReader::buildModel(vector <SSLath*> &faces)
{
  // Faces need not be numbered consecutively, so store here while building.
  map<int, SSLath*> facestmp;
  map<int, vector<int> >::const_iterator cit1, cit2;
  vector<int>::const_iterator cit3;
  map<int, SSVertex>::const_iterator cit4;
  vector<pair<int,int> >::const_iterator cit5;
  map<int, SSLath*>::const_iterator cit6;
  map<int, vector<double> >::const_iterator cit7;
  vector<double>::const_iterator cit8;
  vector<SSVertex> v;
  vector<int> vi;
  pair<int, int> conn;
  SSLath *base, *tmp;
  
  // Build faces
  cit1 = facedata.begin();
  while(cit1 != facedata.end()) 
    {
      vi = cit1->second;
      cit3 = vi.begin();
      while(cit3 != vi.end()) 
        {
          cit4 = vertices.find(*cit3);
          if(cit4 == vertices.end()) 
            {
              ostrstream ost;
              ost << "Invalid vertex " << *cit3 << " used in face " << (cit1->first) << "\n";
              error(ost.str(), 0);
            }
          v.push_back(cit4->second);
          cit3++;
        }
      facestmp.insert(make_pair(cit1->first,SSQModel::Instance()->BuildFace(v,faces)));
      v.clear();
      cit1++;
    }
  
  // Connect faces
  cit5 = connections.begin();
  while(cit5 != connections.end()) 
    {
      cit1 = facedata.find(cit5->first);
      cit2 = facedata.find(cit5->second);
      if(cit1 == facedata.end() || cit2 == facedata.end()) 
        {
          ostrstream ost;
          ost << "Invalid face " << cit5->first << " or " << cit5->second 
              << " used in connection\n";
          error(ost.str(), 0);
        }
      conn = findEdgeNo(cit1->second, cit2->second, cit5->first, cit5->second);
      SSQModel::Instance()->ConnectFaces(facestmp.find(cit5->first)->second, conn.first,
                                         facestmp.find(cit5->second)->second, conn.second);
      cit5++;
    }
  
  // Add knot spacings
  cit7 = knotspacings.begin();
  while(cit7 != knotspacings.end()) {
    cit8 = cit7->second.begin();
    cit6 = facestmp.find(cit7->first);
    if(cit6 == facestmp.end()) {
      ostrstream ost;
      ost << "Invalid face " << cit7->first << " specified for knot spacings.\n";
      error(ost.str(), 0);
    }
    if(2*facedata.find(cit7->first)->second.size() != cit7->second.size()) {
      ostrstream ost;
      ost << "Illegal number of knot spacings specified for face "
          << cit6->first << ".\n"
          << "Two knot spacings must be specified for each edge in the face.\n";
      error(ost.str(), 0);
    }      
    base = cit6->second;
    do {
      base->setcwknotspc(*(cit8++));
      base->setccwknotspc(*(cit8++));
      // Check that knot spacings match on either side of the edge, 
      // and update if existing knot spacings are 1.0.
      tmp = base->ccwv();
      if(tmp && base->cwknotspc() != tmp->ccwknotspc()) {
        if(tmp->ccwknotspc() != 1.0) {
          ostrstream ost;
          ost << "Knot spacings for face " << cit6->first << " do not match knot spacings\n"
              << "in adjacent face across edge " << (cit8-cit7->second.begin()) << ".\n";
          error(ost.str(), 0);
        } else {
          tmp->setccwknotspc(base->cwknotspc());
        }
      }
      tmp = base->cwv();
      if(tmp && base->ccwknotspc() != tmp->cwknotspc()) {
        if(tmp->cwknotspc() != 1.0) {
          ostrstream ost;
          ost << "Knot spacings for face " << cit6->first << " do not match knot spacings\n"
              << "in adjacent face across edge " << (cit8-cit7->second.begin())+1 << ".\n";
          error(ost.str(), 0);
        } else {
          tmp->setcwknotspc(base->ccwknotspc());
        }
      }      
      base = base->ccwf();
    } while(cit8 != cit7->second.end());
    cit7++;
  }
}
  
void SSMdlReader::read(const string &file, vector<SSLath*> &faces)
{
  lineno = 0;
  pushbackflag = 0;
  cout << "Opening file: " << file << endl;
  
  ifstream in(file.c_str());
  if(!in)
    error("Cannot open file "+file+"\n", 0);
  
  while(true) {
    stat = readLine(in);
    
    switch(stat) {
    case IO_EOF:
      goto exit;
    case IO_FAIL:
      error("I/O error. Failed to read.", 1);
    case IO_GOOD:
      break;
    }
#if defined(__SCAN_BUG_WORKAROUND)
    if(!strcmp(buf,"vertices"))
      readVertices(in);
    else if(!strcmp(buf,"faces"))
      readFaces(in);
    else if(!strcmp(buf,"connections"))
      readConnections(in);
    else if(!strcmp(buf,"knotspacings"))
      readKnotSpacings(in);
    else
      error("Illegal file format", 1);
#else
    istrstream is(buf);
    is >> modedata;
    
    if(modedata=="vertices")
      readVertices(in);
    else if(modedata=="faces")
      readFaces(in);
    else if(modedata=="connections")
      readConnections(in);
    else if(modedata=="knotspacings")
      readKnotSpacings(in);
    else
      error("Illegal file format", 1);
#endif
  }
 exit:
  in.close();
  buildModel(faces);
}
