#if defined(WIN32)
#pragma warning( disable : 4786 )
#endif

#include "SSQModel.h"
#include <numeric>
#include <time.h>
#include "SSCornerLath.h"
#include "SSDooSabin.h"
#include "SSCatmullClark.h"
#include "SSLoop.h"
#include "SSQuadraticNURSS.h"
#include "SSCubicNURSS.h"
#include "SSMdlReader.h"
#include "SSVector.h"
#include "SSTypes.h"
#include "SSHeroReader.h"

#if defined(WIN32)
#include <windows.h>
#include <winbase.h>
#else
#include <unistd.h>
#endif

auto_ptr<SSQModel> SSQModel::m_instance = auto_ptr<SSQModel>(0);

SSQModel::SSQModel()
{
  m_method = 0;
  m_curlevel = 0;
}

SSQModel::~SSQModel()
{
  if(m_method)
    delete m_method;

  for(LevelType i = 0; i <= m_curlevel; i++)
    {
      DeleteLevel(i); 
    }

  m_Levels.clear();
}

void SSQModel::DeleteLevel(LevelType l)
{
  if(l < 0 || l > m_curlevel)
    return;

  LathVector laths = m_Levels[l].second;
                
  for(int j = 0; j < laths.size(); j++)
    {
      SSLath *start=laths[j], *tmp=0, *tmp1=0;

      tmp = start;
      int size = 0;
      do
        {
          tmp=tmp->ccwf();
          size++;
        }
      while(tmp != start);

      tmp1 = tmp = start;
      while(size--)
        {
          tmp1 = tmp->ccwf();
                        
          delete tmp;
                        
          tmp = tmp1;
        }
    }
  laths.clear();
}

SSQModel *SSQModel::Instance()
{
  if(m_instance.get() == 0)
    {
      auto_ptr<SSQModel> ptr(new SSQModel());
      m_instance = ptr;
    }
  return m_instance.get();
}

ErrorDescription SSQModel::Init(const string& filename)
{
  vector<SSLath*> v;
  int pos = filename.find_last_of('.');
    
#if defined(WIN32)
  if(pos<filename.size() && !filename.compare(filename.find_last_of('.'), 4, ".mdl"))
#else
    if(pos<filename.size() && !filename.compare(".mdl", filename.find_last_of('.'), 4))
#endif
      {
        SSMdlReader r;
        r.read(filename, v);
        m_curlevel = 0;
        m_Levels.insert(make_pair(m_curlevel, make_pair(v[0], v)));
      }
#if defined(WIN32)
    else if(pos<filename.size() && !filename.compare(filename.find_last_of('.'), 4, ".wrl"))
#else
      else if(pos<filename.size() && !filename.compare(".wrl", filename.find_last_of('.'), 4))
#endif
        {
          SSHeroReader r;
          r.read(filename, v);
          m_curlevel = 0;
          m_Levels.insert(make_pair(m_curlevel, make_pair(v[0], v)));
        } else {
          cerr << "Unknown file format.\nOnly .mdl and .wrl files are supported" << endl;
          exit(1);      
        }
  m_filename = filename;
  return ErrorDescription();
}

SSLath *SSQModel::BuildFace(vector<SSVertex> vertices, vector<SSLath*> &vec)
{

  SSLath *base, *curr, *next, *tmp;
        
  if(vertices.size() < 3 )
    return 0;
  base = new SSCornerLath(vertices[0]);
  curr = base;

  for(int i = 1; i<vertices.size(); i++) 
    {
      next = new SSCornerLath(vertices[i]);
      curr->ccwnextf(next);
      curr = next;
    }

  curr->ccwnextf(base);

  tmp = base;
  int j=0;
  do
    {
      if(vertices[j] != tmp->corner())
        cout << "Error: " << vertices[j] << " / " << tmp->corner() << endl;
      tmp = tmp->ccwf();
      j++;
    }
  while(tmp != base);
  vec.push_back(base);    

  return base;
}

SSLath *SSQModel::BuildFace(const vector<SSVertex> vertices,
                            const vector<KnotSpacingType> knots,
                            vector<SSLath*> &vec)
{
  SSLath *base, *curr, *next;
  int i;

  if(vertices.size() < 3 )
    throw string( "Wrong number of vertices");

  base = new SSCornerLath(vertices[0]);
  curr = base;
        
  for(i = 1; i<vertices.size(); i++) 
    {
      next = new SSCornerLath(vertices[i]);
      curr->ccwnextf(next);
      curr->setcwknotspc(knots[(i-1)*2]);
      curr->setccwknotspc(knots[i*2-1]);
      curr = next;
    }

  curr->ccwnextf(base);
  curr->setcwknotspc(knots[(i-1)*2]);
  curr->setccwknotspc(knots[i*2-1]);
        
  vec.push_back(base);
  return base;
}

ErrorDescription SSQModel::ConnectFaces(SSLath* f1, int edgeno1, SSLath* f2, int edgeno2)
{
  while(edgeno1 > 0) 
    { 
      f1 = f1->ccwf(); 
      edgeno1--; 
    }
        
  while(edgeno2 > 0) 
    { 
      f2 = f2->ccwf(); 
      edgeno2--; 
    }

  f1->ccwnextv(f2->cwf());
  f2->ccwnextv(f1->cwf());
        
  return ErrorDescription();
}

ErrorDescription SSQModel::FaceVertices(vector<SSVertex> &vertices, SSLath* start)
{
  SSLath *tmp;
  
  tmp = start;
  do 
    {
      vertices.push_back(tmp->corner());
      tmp = tmp->ccwf();
    } 
  while(start != tmp);

  return ErrorDescription();
}

ErrorDescription SSQModel::FaceVertices(vector<SSVertex> &vertices, 
                                        vector<KnotSpacingType> &knots,
                                        SSLath* start)
{
  SSLath *tmp;
  
  tmp = start;
  do 
    {
      vertices.push_back(tmp->corner());
      knots.push_back(tmp->cwknotspc());
      knots.push_back(tmp->ccwknotspc());
      tmp = tmp->ccwf();
    } 
  while(start != tmp);

  return ErrorDescription();
}

ErrorDescription SSQModel::SwitchMethod(MethodType newmethod)
{
  if(m_curmethod != newmethod)
    {
      if(m_method)
        delete m_method;

      for(LevelType i = 0; i <= m_curlevel; i++)
        {
          DeleteLevel(i); 
        }

      m_Levels.clear();

      LevelType tmpl = m_curlevel;

      time_t s = clock();
      Init(m_filename);
      time_t e = clock();

      cout << "Init: " << e - s << " msec " << endl;
                
      s = clock();
      ErrorDescription ed = Subdivide(newmethod, tmpl);
      e = clock();

#if defined(WIN32)
      cout << "Subdivide: " << e - s << " msec / levels: " << tmpl << endl;
#else
      cout << "Subdivide: " << ((double)(e - s))/CLOCKS_PER_SEC 
           << " msec / levels: " << tmpl << endl;
#endif
      return ed;
    }
  return ErrorDescription();
}

void SubdivideMethod(int nolevels,  LevelType &curlevel, 
                     map<LevelType, Level> &Levels, SSSubdivideMethods *method)
{
  vector<SSLath*> v;
  SSLath *levelstart=0;

  levelstart = Levels[curlevel].first;
        
  for(int i = 1; i <= nolevels; i++)
    {
      time_t start = clock();
      levelstart = method->Subdivision(v, levelstart);
        
      // Subdivide unconnected parts
      vector<SSLath*> faces = Levels[curlevel].second;
      for(int q=0; q<faces.size(); q++) 
        {
          if(!faces[q]->nl())
            method->Subdivision(v, faces[q]);
        }
                
      time_t stop = clock();
      cout << "Level "<<i << "  Polygons per sec: " 
           << ((double)v.size())/(((double)(stop-start))/((double)CLOCKS_PER_SEC)) << endl;
                
      curlevel++;
      Levels.insert(make_pair(curlevel, make_pair(levelstart, v)));
      v.clear();
    }
}

ErrorDescription SSQModel::Subdivide(MethodType m, int levels)
{
  try
    {
      vector<SSLath*> v;
      SSLath *levelstart=0;

      m_curmethod = m;

      switch(m)
        {
        case c_DooSabin:
          m_method = new SSDooSabin();
          break;

        case c_CatmullClark:
          m_method = new SSCatmullClark();
          break;

        case c_Loop:
          m_method = new SSLoop();
#ifdef TRIANGULATE
          Triangulate();
#endif
          break;

        case c_QuadNURSS:
          m_method = new SSQuadraticNURSS();
          break;

        case c_CubicNURSS:
          m_method = new SSCubicNURSS();
          break;
        }
      SubdivideMethod(levels, m_curlevel, m_Levels, m_method);
    } catch(...)
      {
        cout << "Subdiv error" << endl;
        return ErrorDescription(3000, "Subdivision Error", "");
      }
  return ErrorDescription();
}

ErrorDescription SSQModel::Draw(int level, int flags)
{
  if(level < 0 || level > m_curlevel)
    level = 0;

  LathVector laths = m_Levels[level].second;
  LathVector::iterator viter = laths.begin();
                        
  while(viter != laths.end())
    {
      DrawFace(*viter, flags);
                  
      if(flags & SS_DRAW_TRACE)
        {
          glutSwapBuffers();
#if defined(WIN32)
          Sleep(30 / laths.size());
#else
          usleep(30000 / laths.size());
#endif
        }
      viter++;
    }
    
  return ErrorDescription();
}

SSVector SSQModel::calcNormal(SSLath *face)
{
  SSVector norm(0.0, 0.0, 0.0);
  SSLath *tmp;
  tmp = face;
  int count = 0;
  do {
    SSVector v1(tmp->corner(), tmp->ccwf()->corner());
    SSVector v2(tmp->corner(), tmp->cwf()->corner());
    v1.normalize(); v2.normalize();
    SSVector c = cross(v1, v2);
    norm += c;
    tmp = tmp->ccwv();
    count++;
    if(count > 200) 
      { // Possible infinite loop, most likely because of wrong connections
        return SSVector(1.0,0,0);  
      }
  } while(tmp && tmp != face);
  norm.normalize();
  return norm;
}

void SSQModel::DrawFace(SSLath *start, int flags)
{
  SSLath *l1;
  l1 = start;
  SSVector c;
  SSVertex midpoint;
        
  if(flags & SS_DRAW_OUTLINE) 
    {
      do
        {
          glBegin(GL_LINES);
              
          midpoint = 0.5 * (l1->corner() + l1->ccwf()->corner());
              
          if(l1->ccwknotspc() == 0.0)
            glColor3f(1.0,0.0,0.0);
          else if(l1->ccwknotspc() == 2.0)
            glColor3f(1.0,1.0,0.0);
          else if(l1->ccwknotspc() == 3.0)
            glColor3f(0.0,1.0,0.0);
          else if(l1->ccwknotspc() == 4.0)
            glColor3f(0.0,0.0,1.0);
          else
            glColor3f(1.0,1.0,1.0);
              
          glVertex3f((GLfloat) l1->corner().x(),
                     (GLfloat) l1->corner().y(),
                     (GLfloat) l1->corner().z());
          glVertex3f((GLfloat) midpoint.x(),
                     (GLfloat) midpoint.y(),
                     (GLfloat) midpoint.z());
              
          l1 = l1->ccwf();
              
          if(l1->cwknotspc() == 0.0)
            glColor3f(1.0,0.0,0.0);
          else if(l1->cwknotspc() == 2.0)
            glColor3f(1.0,1.0,0.0);
          else if(l1->cwknotspc() == 3.0)
            glColor3f(0.0,1.0,0.0);
          else if(l1->cwknotspc() == 4.0)
            glColor3f(0.0,0.0,1.0);
          else
            glColor3f(1.0,1.0,1.0);
              
          glVertex3f((GLfloat) l1->corner().x(),
                     (GLfloat) l1->corner().y(),
                     (GLfloat) l1->corner().z());
          glVertex3f((GLfloat) midpoint.x(),
                     (GLfloat) midpoint.y(),
                     (GLfloat) midpoint.z());
              
        } 
      while(l1 != start);
      glEnd();
    } 
  else 
    {      
      glBegin(GL_POLYGON);
      l1 = start;
    
      do 
        {
          if(flags & SS_DRAW_EXACT_NORMALS)
            c = m_method->calcNormal(l1);
          else
            c = calcNormal(l1);
       
          glNormal3f(c[0], c[1], c[2]);
        
          glVertex3f((GLfloat) l1->corner().x(),
                     (GLfloat) l1->corner().y(),
                     (GLfloat) l1->corner().z());
        
        
          l1 = l1->ccwf();
        
        } 
      while(l1 != start);
    
      glEnd();
    }
}

void SSQModel::Triangulate()
{
  int i,j;
  SSLath *face;
  LathVector laths = m_Levels[0].second;
  LathVector newlaths;
  for(i=0; i<laths.size(); i++)
    {
      face = laths[i];
      vector<SSVertex> v;
      FaceVertices(v, face);
      SSVertex center = accumulate(v.begin(), v.end(), SSVertex())/v.size();
      SSLath *base, *curr, *next, *saved, *nextbase;
      base = new SSCornerLath(center);
      curr = base;
      for(j=0; j<v.size()-1; j++)
        {
          next = new SSCornerLath(center);
          newlaths.push_back(next);
          curr->ccwnextv(next);
          curr = next;
        }
      next->ccwnextv(base);
      newlaths.push_back(base);

      curr = face;
      do {
        next = new SSCornerLath(curr->corner());
        next->ccwnextv(curr->ccwv());
        curr->ccwnextv(next);
        // Maintain knot spacings
        next->setcwknotspc(curr->cwknotspc());
        next->setccwknotspc(1.0);
        curr->setcwknotspc(1.0);

        curr = curr->ccwf();
      }
      while(curr != face);
      
      curr = face;
      nextbase = base;
      do {
        saved = curr->ccwf();
        curr->ccwnextf(saved->ccwv());
        curr->ccwf()->ccwnextf(nextbase);
        nextbase->ccwnextf(curr);
        nextbase = nextbase->ccwv();
        curr = saved;
      }
      while(curr != face);

    }
  m_Levels[0] = make_pair(newlaths[0], newlaths);
}
