#include <cmath>
#include "SSQuadraticNURSS.h"
#include "SSQModel.h"
#include "SSCornerLath.h"

SSQuadraticNURSS::SSQuadraticNURSS()
{
}

SSQuadraticNURSS::~SSQuadraticNURSS()
{
}

SSLath *SSQuadraticNURSS::Subdivision(vector<SSLath*> &vec, SSLath *start)
{
  SSQuadraticNURSS::Subdivide(vec, start);
  return start->nl();
}

void SSQuadraticNURSS::Subdivide(vector<SSLath*> &vec, SSLath *start)
{
  // If null or already subdivided just return
  if(!start || start->nl())
    return;


  vector<SSVertex> vs, evs(4, SSVertex()), nvs;
  vector<SSLath *> els(4, (SSLath *) 0);
  vector<CoefficientType> coeff;
  vector<KnotSpacingType> knots, nknots;
        
  SSLath *nf=0, *tmp1=0, *tmp2=0, *tmp3=0;

  int i, j, c, k;
  SSVertex V, S, ny;
  KnotSpacingType denom = 0, t;

        
  SSQModel::Instance()->FaceVertices(vs, knots, start);
  k = vs.size();
        
  // Calculate common coefficients
  tmp1 = start;
  for(i=0; i<k; i++)
    {
      t = tmp1->knotspc(-1,0,0) * tmp1->knotspc(1,0,0);
      denom += t;
      V += vs[i]*t;
      coeff.push_back(1.0+2.0*cos(2.0*((double) i)*c_PI/((double) k)));
      tmp1 = tmp1->ccwf();
    }
  if(abs(denom) >= 1e-10)
    V /= denom;


  tmp1 = start;
  for(i=0; i<k; i++) 
    {
      S = SSVertex();
      for(j=0; j<k; j++) 
        {
          c = (i+j)%k;
          S += coeff[j] * vs[c];
        }
      if(abs(denom) < 1e-10)
        // If denominator zero -> point is not changed
        nvs.push_back(vs[i]);
      else {
        // Sederberg et al. eq (10)
        t = tmp1->knotspc(1,2,0) * tmp1->knotspc(3,2,0)
          + tmp1->knotspc(-1,-2,0) * tmp1->knotspc(-3,-2,0);
                    
        ny = 0.5*(V+vs[i]) + t * (-k * vs[i] + S)/(8.0*denom);
        nvs.push_back(ny);
      }
      tmp1 = tmp1->ccwf();
    }

  // Calculate new knot spacings
  for(i=0; i<k; i++)
    {
      nknots.push_back(knots[2*i]*0.5);
      nknots.push_back(knots[2*i+1]*0.5);
    }

  nf = SSQModel::Instance()->BuildFace(nvs, nknots, vec);
        
  // Connect to subdivided level
  start->nextlevel(nf);
        
  tmp1 = start;
  tmp2 = nf;
  do 
    {
      tmp1->nextlevel(tmp2);
      tmp1 = tmp1->ccwf();
      tmp2 = tmp2->ccwf();
    } 
  while(tmp1 != start);
        
  // Start subdivision chain reaction
  tmp1 = start;
  do 
    {
      SSQuadraticNURSS::Subdivide(vec, tmp1->ccwv());
      tmp1 = tmp1->ccwf();
    } 
  while(start != tmp1);

  // Build and connect faces corresponding to old edges
  tmp1 = start;
  do
    {
      nknots = vector<KnotSpacingType>(8,0);
      if(tmp1->ccwv() && !tmp1->nl()->ccwv()) 
        {
          tmp2 = tmp1;
          evs[3] = tmp2->nl()->corner();
          els[0] = tmp2->nl();
          nknots[6] = 0.25 * (tmp2->knotspc(0,-1,0) + tmp2->knotspc(0,1,1));
          nknots[7] = 0.5 * tmp2->knotspc(0,1,0);
          tmp2 = tmp2->ccwv();
          evs[2] = tmp2->nl()->corner();
          nknots[4] = 0.5 * tmp2->knotspc(0,-1,0);
          nknots[5] = nknots[6];
          tmp2 = tmp2->ccwf();
          evs[1] = tmp2->nl()->corner();
          els[1] = tmp2->nl();
          nknots[2] = 0.25 * (tmp2->knotspc(0,-1,0) + tmp2->knotspc(0,1,1));
          nknots[3] = 0.5 * tmp2->knotspc(0,1,0);
          tmp2 = tmp2->ccwv();
          evs[0] = tmp2->nl()->corner();
          nknots[0] = 0.5 * tmp2->knotspc(0,-1,0);
          nknots[1] = nknots[2];
              
          tmp2 = SSQModel::Instance()->BuildFace(evs, nknots, vec);
              
          SSQModel::Instance()->ConnectFaces(els[0], 0, tmp2, 0);
          SSQModel::Instance()->ConnectFaces(els[1], 0, tmp2, 2);
        }
      tmp1 = tmp1->ccwf();
    }
  while(tmp1 != start);

  // Build and connect faces corresponding to old vertices
  tmp1 = start;
          
  do 
    {
      if(tmp1->ccwv() && !tmp1->nl()->ccwv()->ccwv())
        { // If not already connected, build & connect face
          tmp2 = tmp1;
          vs.clear();
          nknots.clear();
          do 
            {
              tmp3 = tmp2->nl();
              if(!tmp3 || !tmp3->ccwv()) 
                {
                  /* All faces sharing edges with f have been subdivided in the
                   * recursion, but faces sharing only vertices with f may not
                   * have been.  Solution: Just skip affected vertex -- it will
                   * be processed further down the recursion.
                   * Unknown why second condition necessary!!!
                   **/
                  goto loop;
                }
              vs.push_back(tmp3->corner());
              nknots.push_back(tmp3->cwv()->ccwknotspc());
              nknots.push_back(tmp3->ccwv()->cwknotspc());
                 
              tmp2 = tmp2->ccwv();
                  
              if(!tmp2)
                goto loop;
            } 
          while(tmp1  != tmp2);
                
          tmp3 = SSQModel::Instance()->BuildFace(vs, nknots, vec);
                  
          do 
            {
              if(tmp2->nl()->ccwv())
                SSQModel::Instance()->ConnectFaces(tmp2->nl()->ccwv(), 0, tmp3, 1);
                  
              tmp2 = tmp2->ccwv();
              tmp3 = tmp3->ccwf();
            } 
          while(tmp1 != tmp2);
          
        }
    loop:
      tmp1 = tmp1->ccwf();
    }
  while(tmp1 != start);
}
