#ifndef _SSSSVector_HEADER
#define _SSSSVector_HEADER

#include "SSMaster.h"

#include <stdlib.h>
#include <math.h>

#include "SSVertex.h"

// Class representing a vector in homogeneous coordinates
class SSVector 
{

public:  
  // Constructors
  SSVector(SSVertex f, SSVertex t);
  SSVector(SSVertex f);
  SSVector( const double _a, const double _b, const double _c );
  SSVector( const double ar[3]);
  SSVector( const double _a, const double _b, const double _c, const double _d );
  SSVector(void);

  // Non-destructive arithmetic operators
  SSVector operator + ( const SSVector &x ) const;
  SSVector operator - ( const SSVector &x ) const;
  SSVector operator - ( void ) const;
  SSVector operator * ( const double &k ) const;

  // Destructive arithmetic operators
  void operator -= ( const SSVector &x );
  void operator += ( const SSVector &x );
  void operator *= ( const SSVector &x );
  void operator *= ( const double &k );
  void operator /= ( const double &k );

  // Comparison operators
  int operator < ( const SSVector &x ) const; 
  int operator <= ( const SSVector &x ) const;
  int operator > ( const SSVector &x ) const;
  int operator >= ( const SSVector &x ) const;
  int operator != ( const int &val ) const;
  int operator != ( const SSVector &v ) const;
  int operator == ( const SSVector &v ) const;

  // Cross product, Euclidean length
  void cross( const SSVector &x, const SSVector &y );
  double len( void ) const;

  // Normalization and homogenization
  void normalize( void );
  void homogenize()
  {
    a = a/d;
    b = b/d;
    c = c/d;
    d = 1;
  };

  // Reverse direction of vector
  void invert();

  // Make vector x orthogonal to projection of current vector onto XY-plane
  void get_ortogonal( SSVector &x ) const;

  // Indexing and assignment
  void copy_to_array( double dest[3] ) const;
  double& operator [] ( const int index );
  double operator [] ( const int index ) const;
  void operator = (const SSVector &v);

private:
  double a,b,c,d;

  // Non-destructive dot and cross product
  friend inline double dot( const SSVector &x, const SSVector &y );
  friend inline double dot( const SSVector &x, const double v[3] );
  friend inline SSVector cross( const SSVector &x, const SSVector &y );  
};

// Constant null vector
const SSVector NULLSSVector;

inline SSVector::SSVector(SSVertex f) 
{
  a = f.x();
  b = f.y();
  c = f.z();
  d = 1.0;
}

inline SSVector::SSVector(SSVertex f, SSVertex t) 
{
  a = f.x() - t.x();
  b = f.y() - t.y();
  c = f.z() - t.z();
  d = 1.0;
}

inline SSVector::SSVector(void) 
{
  a=b=c=0.0; 
  d=1.0;
}

inline SSVector::SSVector( const double _a, const double _b, const double _c ):
  a(_a),b(_b),c(_c)
{
  d = 1.0;
}

inline SSVector::SSVector( const double ar[3]) : a(ar[0]), b(ar[1]), c(ar[2]) 
{
  d = 1.0;
}

inline SSVector::SSVector( const double _a, const double _b, 
                           const double _c, const double _d):
  a(_a),b(_b),c(_c),d(_d) {}

inline SSVector SSVector::operator + ( const SSVector &x ) const 
{
  return SSVector( x.a+a, x.b+b, x.c+c);
}

inline SSVector SSVector::operator - ( const SSVector &x ) const 
{
  return SSVector( a-x.a, b-x.b, c-x.c );
}

inline SSVector SSVector::operator - ( void ) const 
{
  return SSVector( -a, -b, -c ); 
}

inline SSVector SSVector::operator * ( const double &k ) const 
{
  return SSVector( a*k, b*k, c*k );
}

inline void SSVector::operator -= ( const SSVector &x ) 
{ 
  a-=x.a; 
  b-=x.b; 
  c-=x.c; 
}

inline void SSVector::operator += ( const SSVector &x ) 
{
  a+=x.a; 
  b+=x.b; 
  c+=x.c; 
}

inline void SSVector::operator *= ( const SSVector &x ) 
{ 
  a*=x.a; 
  b*=x.b; 
  c*=x.c; 
}

inline void SSVector::operator *= ( const double &k ) 
{ 
  a*=k; 
  b*=k; 
  c*=k; 
}

inline void SSVector::operator /= ( const double &k ) 
{ 
  a/=k; 
  b/=k; 
  c/=k; 
}

inline int SSVector::operator < ( const SSVector &x ) const 
{
  return (a<x.a && b<x.b && c<x.c);
}

inline int SSVector::operator <= ( const SSVector &x ) const 
{
  return (a<=x.a && b<=x.b && c<=x.c);
}

inline int SSVector::operator > ( const SSVector &x ) const 
{
  return (a>x.a && b>x.b && c>x.c);
}

inline int SSVector::operator >= ( const SSVector &x ) const 
{
  return (a>=x.a && b>=x.b && c>=x.c);
}

inline int SSVector::operator != ( const int &val ) const 
{
  if (val==0 && a==0 && b==0 && c==0)
    return 0;
  else
    return 1;
}

inline int SSVector::operator != ( const SSVector &v ) const 
{
  if (a!=v.a || b!=v.b || c!= v.c)
    return 1;
  else
    return 0;
}

inline int SSVector::operator == ( const SSVector &v ) const 
{
  if (a==v.a && b==v.b && c== v.c)
    return 1;
  else
    return 0;
}

inline void SSVector::cross( const SSVector &x, const SSVector &y ) 
{
  a = x.b*y.c-x.c*y.b;
  b = x.c*y.a-x.a*y.c;
  c = x.a*y.b-x.b*y.a;
  d = 1;
}

inline double SSVector::len( void ) const 
{ 
  return sqrt( a*a + b*b + c*c ); 
}

inline void SSVector::normalize( void ) 
{
  double len;
  if ((len = sqrt(a*a+b*b+c*c)) == 0.0)
    {
      cerr << "Can't normalize nullSSVector! Using [1,0,0]." << endl;
      a=1.0; b=0.0; c=0.0;
      return;
    }
  double invlen = 1.0/len;
  a *= invlen; b *= invlen; c *= invlen;
}

inline void SSVector::invert() 
{ 
  a = -a; 
  b = -b; 
  c = -c; 
}

inline void SSVector::copy_to_array( double dest[3] ) const 
{
  dest[0] = a; dest[1] = b; dest[2] = c;
}

inline double& SSVector::operator [] ( const int index )
{
  switch( index ) {
  case 0: return a;
  case 1: return b;
  case 2: return c;
  case 3: return d;
  default:
    cerr <<  "SSVector index must be in [0,3]";
    exit(1);
  }
  return a; // To avoid warnings - we don't get here
}

inline double SSVector::operator [] ( const int index ) const
{
  switch( index ) 
    {
    case 0: return a;
    case 1: return b;
    case 2: return c;
    case 3: return d;
    default:
      cerr << "SSVector index must be in [0,3]";
      exit(1);
    }
  return 0.0;
}

inline void SSVector::operator = (const SSVector &v)
{
  a = v.a; b = v.b; c = v.c;
}

inline double dot( const SSVector &x, const SSVector &y ) 
{
  return x.a*y.a + x.b*y.b + x.c*y.c;
}

inline double dot( const SSVector &x, const double v[3] ) 
{
  return x.a*v[0] + x.b*v[1] + x.c*v[2];
}

inline SSVector cross( const SSVector &x, const SSVector &y ) 
{
  return SSVector( x.b*y.c-x.c*y.b, x.c*y.a-x.a*y.c, x.a*y.b-x.b*y.a );
}

ostream& operator<<(ostream& os, const SSVector& x );

#endif
